"""
Module for Firing Events via PagerDuty

.. versionadded:: 2014.1.0

:configuration: This module can be used by specifying the name of a
    configuration profile in the minion config, minion pillar, or master
    config.

    For example:

    .. code-block:: yaml

        my-pagerduty-account:
            pagerduty.api_key: F3Rbyjbve43rfFWf2214
            pagerduty.subdomain: mysubdomain
"""

import salt.utils.functools
import salt.utils.json
import salt.utils.pagerduty


def __virtual__():
    """
    No dependencies outside of what Salt itself requires
    """
    return True


def list_services(profile=None, api_key=None):
    """
    List services belonging to this account

    CLI Example:

    .. code-block:: bash

        salt myminion pagerduty.list_services my-pagerduty-account
    """
    return salt.utils.pagerduty.list_items(
        "services", "name", __salt__["config.option"](profile), api_key, opts=__opts__
    )


def list_incidents(profile=None, api_key=None):
    """
    List incidents belonging to this account

    CLI Example:

    .. code-block:: bash

        salt myminion pagerduty.list_incidents my-pagerduty-account
    """
    return salt.utils.pagerduty.list_items(
        "incidents", "id", __salt__["config.option"](profile), api_key, opts=__opts__
    )


def list_users(profile=None, api_key=None):
    """
    List users belonging to this account

    CLI Example:

    .. code-block:: bash

        salt myminion pagerduty.list_users my-pagerduty-account
    """
    return salt.utils.pagerduty.list_items(
        "users", "id", __salt__["config.option"](profile), api_key, opts=__opts__
    )


def list_schedules(profile=None, api_key=None):
    """
    List schedules belonging to this account

    CLI Example:

    .. code-block:: bash

        salt myminion pagerduty.list_schedules my-pagerduty-account
    """
    return salt.utils.pagerduty.list_items(
        "schedules", "id", __salt__["config.option"](profile), api_key, opts=__opts__
    )


def list_windows(profile=None, api_key=None):
    """
    List maintenance windows belonging to this account

    CLI Example:

    .. code-block:: bash

        salt myminion pagerduty.list_windows my-pagerduty-account
        salt myminion pagerduty.list_maintenance_windows my-pagerduty-account
    """
    return salt.utils.pagerduty.list_items(
        "maintenance_windows",
        "id",
        __salt__["config.option"](profile),
        api_key,
        opts=__opts__,
    )


# The long version, added for consistency
list_maintenance_windows = salt.utils.functools.alias_function(
    list_windows, "list_maintenance_windows"
)


def list_policies(profile=None, api_key=None):
    """
    List escalation policies belonging to this account

    CLI Example:

    .. code-block:: bash

        salt myminion pagerduty.list_policies my-pagerduty-account
        salt myminion pagerduty.list_escalation_policies my-pagerduty-account
    """
    return salt.utils.pagerduty.list_items(
        "escalation_policies",
        "id",
        __salt__["config.option"](profile),
        api_key,
        opts=__opts__,
    )


# The long version, added for consistency
list_escalation_policies = salt.utils.functools.alias_function(
    list_policies, "list_escalation_policies"
)


def create_event(
    service_key=None, description=None, details=None, incident_key=None, profile=None
):
    """
    Create an event in PagerDuty. Designed for use in states.

    CLI Example:

    .. code-block:: yaml

        salt myminion pagerduty.create_event <service_key> <description> <details> \
        profile=my-pagerduty-account

    The following parameters are required:

    service_key
        This key can be found by using pagerduty.list_services.

    description
        This is a short description of the event.

    details
        This can be a more detailed description of the event.

    profile
        This refers to the configuration profile to use to connect to the
        PagerDuty service.
    """
    trigger_url = "https://events.pagerduty.com/generic/2010-04-15/create_event.json"

    if isinstance(details, str):
        details = salt.utils.yaml.safe_load(details)
        if isinstance(details, str):
            details = {"details": details}

    ret = salt.utils.json.loads(
        salt.utils.pagerduty.query(
            method="POST",
            profile_dict=__salt__["config.option"](profile),
            api_key=service_key,
            data={
                "service_key": service_key,
                "incident_key": incident_key,
                "event_type": "trigger",
                "description": description,
                "details": details,
            },
            url=trigger_url,
            opts=__opts__,
        )
    )
    return ret



"""
Manage Microsoft Store apps on Windows. Removing an app with this modules will
deprovision the app from the online Windows image.

.. versionadded:: 3007.0
"""

import fnmatch
import logging

import salt.utils.data
import salt.utils.platform

log = logging.getLogger(__name__)
__virtualname__ = "appx"


def __virtual__():
    """
    Only work on Windows where the DISM module is available
    """
    if not salt.utils.platform.is_windows():
        return False, "Appx state: Only available on Windows"

    pwsh_info = __salt__["cmd.shell_info"](shell="powershell", list_modules=False)
    if not pwsh_info["installed"]:
        return False, "Appx state: PowerShell not available"

    return __virtualname__


def absent(name, query, include_store=False, frameworks=False, deprovision_only=False):
    """
    Removes Microsoft Store packages from the system. If the package is part of
    a bundle, the entire bundle will be removed.

    This function removes the package for all users on the system. It also
    deprovisions the package so that it isn't re-installed by later system
    updates. To only deprovision a package and not remove it for all users, set
    ``deprovision_only=True``.

    Args:

        query (str):
            The query string to use to select the packages to be removed. If the
            string matches multiple packages, they will all be removed. Here are
            some example strings:

            | string          | description |
            | --------------- | ----------- |
            | ``*teams*``     | Remove Microsoft Teams |
            | ``*zune*``      | Remove Windows Media Player and Zune Video |
            | ``*zuneMusic*`` | Only remove Windows Media Player |
            | ``*xbox*``      | Remove all xBox packages, there are 5 by default
            | ``*``           | Remove everything but the Microsoft Store, unless ``include_store=True`` |

            .. note::
                Use the ``appx.list`` function to make sure your query is
                returning what you expect. Then use the same query to remove
                those packages

        include_store (bool):
            Include the Microsoft Store in the results of the query to be
            removed. Use this with caution. It is difficult to reinstall the
            Microsoft Store once it has been removed with this function. Default
            is ``False``

        frameworks (bool):
            Include frameworks in the results of the query to be removed.
            Default is ``False``

        deprovision_only (bool):
            Only deprovision the package. The package will be removed from the
            current user and added to the list of deprovisioned packages. The
            package will not be re-installed in future system updates. New users
            of the system will not have the package installed. However, the
            package will still be installed for existing users. Default is
            ``False``

    Returns:
        bool: ``True`` if successful, ``None`` if no packages found

    Raises:
        CommandExecutionError: On errors encountered removing the package

    CLI Example:

    .. code-block:: yaml

        remove_candy_crush:
          appx.absent:
            - query: "*candy*"
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    old = __salt__["appx.list"](include_store=include_store, frameworks=frameworks)
    matches = fnmatch.filter(old, query)
    if not matches:
        ret["comment"] = f"No apps found matching query: {query}"
        return ret

    if __opts__["test"]:
        comment = ["The following apps will be removed:"]
        comment.extend(matches)
        ret["comment"] = "\n- ".join(comment)
        ret["result"] = None
        return ret

    # Install the capability
    status = __salt__["appx.remove"](
        query,
        include_store=include_store,
        frameworks=frameworks,
        deprovision_only=deprovision_only,
    )

    if status is None:
        ret["comment"] = f"No apps found matching query: {query}"
        ret["result"] = False

    new = __salt__["appx.list"](include_store=include_store, frameworks=frameworks)
    changes = salt.utils.data.compare_lists(old, new)

    if changes:
        ret["comment"] = f"Removed apps matching query: {query}"
        ret["changes"] = changes

    return ret



"""
Manage launchd plist files
"""

import os
import sys


def write_launchd_plist(program):
    """
    Write a launchd plist for managing salt-master or salt-minion

    CLI Example:

    .. code-block:: bash

        salt-run launchd.write_launchd_plist salt-master
    """
    plist_sample_text = """
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
  <dict>
    <key>Label</key>
    <string>org.saltstack.{program}</string>
    <key>RunAtLoad</key>
    <true/>
    <key>KeepAlive</key>
    <true/>
    <key>ProgramArguments</key>
    <array>
        <string>{script}</string>
    </array>
    <key>SoftResourceLimits</key>
    <dict>
        <key>NumberOfFiles</key>
        <integer>100000</integer>
    </dict>
    <key>HardResourceLimits</key>
    <dict>
        <key>NumberOfFiles</key>
        <integer>100000</integer>
    </dict>
  </dict>
</plist>
    """.strip()

    supported_programs = ["salt-master", "salt-minion"]

    if program not in supported_programs:
        sys.stderr.write(f"Supported programs: '{supported_programs}'\n")
        sys.exit(-1)

    return plist_sample_text.format(
        program=program,
        python=sys.executable,
        script=os.path.join(os.path.dirname(sys.executable), program),
    )



import time

import pytest

import salt.netapi.rest_cherrypy.app as cherrypy_app
from tests.support.mock import MagicMock, patch


class MockCherryPy:
    session = MagicMock(cache={})
    config = {"saltopts": {}}


class MockResolver:
    def __init__(self, *args, **kwargs):
        pass

    def get_token(self, token):
        pass


@pytest.fixture
def configure_loader_modules():
    return {cherrypy_app: {}}


def test__is_valid_token():
    with patch("salt.netapi.rest_cherrypy.app.cherrypy", MockCherryPy()):
        with patch("salt.auth.Resolver", MockResolver):
            events = cherrypy_app.Events()
            with patch.object(
                events.resolver, "get_token", return_value={"expire": time.time() + 60}
            ):
                assert events._is_valid_token("ABCDEF")


def test__is_valid_token_expired():
    with patch("salt.netapi.rest_cherrypy.app.cherrypy", MockCherryPy()):
        with patch("salt.auth.Resolver", MockResolver):
            events = cherrypy_app.Events()
            with patch.object(
                events.resolver, "get_token", return_value={"expire": time.time() - 60}
            ):
                assert not events._is_valid_token("ABCDEF")



import pytest

import salt.config
import salt.daemons.masterapi as masterapi
import salt.utils.platform
from tests.support.mock import MagicMock, patch

pytestmark = [
    pytest.mark.slow_test,
]


class FakeCache:
    def __init__(self):
        self.data = {}

    def store(self, bank, key, value):
        self.data[bank, key] = value

    def fetch(self, bank, key):
        return self.data[bank, key]


@pytest.fixture
def funcs(temp_salt_master):
    opts = temp_salt_master.config.copy()
    salt.cache.MemCache.data.clear()
    funcs = masterapi.RemoteFuncs(opts)
    funcs.cache = FakeCache()
    return funcs


def test_mine_get(funcs, tgt_type_key="tgt_type"):
    """
    Asserts that ``mine_get`` gives the expected results.

    Actually this only tests that:

    - the correct check minions method is called
    - the correct cache key is subsequently used
    """
    funcs.cache.store("minions/webserver", "mine", dict(ip_addr="2001:db8::1:3"))
    with patch(
        "salt.utils.minions.CkMinions._check_compound_minions",
        MagicMock(return_value=dict(minions=["webserver"], missing=[])),
    ):
        ret = funcs._mine_get(
            {
                "id": "requester_minion",
                "tgt": "G@roles:web",
                "fun": "ip_addr",
                tgt_type_key: "compound",
            }
        )
    assert ret == dict(webserver="2001:db8::1:3")


def test_mine_get_pre_nitrogen_compat(funcs):
    """
    Asserts that pre-Nitrogen API key ``expr_form`` is still accepted.

    This is what minions before Nitrogen would issue.
    """
    test_mine_get(funcs, tgt_type_key="expr_form")


def test_mine_get_dict_str(funcs, tgt_type_key="tgt_type"):
    """
    Asserts that ``mine_get`` gives the expected results when request
    is a comma-separated list.

    Actually this only tests that:

    - the correct check minions method is called
    - the correct cache key is subsequently used
    """
    funcs.cache.store(
        "minions/webserver",
        "mine",
        dict(ip_addr="2001:db8::1:3", ip4_addr="127.0.0.1"),
    )
    with patch(
        "salt.utils.minions.CkMinions._check_compound_minions",
        MagicMock(return_value=dict(minions=["webserver"], missing=[])),
    ):
        ret = funcs._mine_get(
            {
                "id": "requester_minion",
                "tgt": "G@roles:web",
                "fun": "ip_addr,ip4_addr",
                tgt_type_key: "compound",
            }
        )
    assert ret == dict(
        ip_addr=dict(webserver="2001:db8::1:3"),
        ip4_addr=dict(webserver="127.0.0.1"),
    )


def test_mine_get_dict_list(funcs, tgt_type_key="tgt_type"):
    """
    Asserts that ``mine_get`` gives the expected results when request
    is a list.

    Actually this only tests that:

    - the correct check minions method is called
    - the correct cache key is subsequently used
    """
    funcs.cache.store(
        "minions/webserver",
        "mine",
        dict(ip_addr="2001:db8::1:3", ip4_addr="127.0.0.1"),
    )
    with patch(
        "salt.utils.minions.CkMinions._check_compound_minions",
        MagicMock(return_value=dict(minions=["webserver"], missing=[])),
    ):
        ret = funcs._mine_get(
            {
                "id": "requester_minion",
                "tgt": "G@roles:web",
                "fun": ["ip_addr", "ip4_addr"],
                tgt_type_key: "compound",
            }
        )
    assert ret == dict(
        ip_addr=dict(webserver="2001:db8::1:3"),
        ip4_addr=dict(webserver="127.0.0.1"),
    )


def test_mine_get_acl_allowed(funcs):
    """
    Asserts that ``mine_get`` gives the expected results when this is allowed
    in the client-side ACL that was stored in the mine data.
    """
    funcs.cache.store(
        "minions/webserver",
        "mine",
        {
            "ip_addr": {
                salt.utils.mine.MINE_ITEM_ACL_DATA: "2001:db8::1:4",
                salt.utils.mine.MINE_ITEM_ACL_ID: salt.utils.mine.MINE_ITEM_ACL_VERSION,
                "allow_tgt": "requester_minion",
                "allow_tgt_type": "glob",
            },
        },
    )
    # The glob check is for the resolution of the allow_tgt
    # The compound check is for the resolution of the tgt in the mine_get request.
    with patch(
        "salt.utils.minions.CkMinions._check_glob_minions",
        MagicMock(return_value={"minions": ["requester_minion"], "missing": []}),
    ), patch(
        "salt.utils.minions.CkMinions._check_compound_minions",
        MagicMock(return_value={"minions": ["webserver"], "missing": []}),
    ):
        ret = funcs._mine_get(
            {
                "id": "requester_minion",
                "tgt": "anything",
                "tgt_type": "compound",
                "fun": ["ip_addr"],
            }
        )
    assert ret == {"ip_addr": {"webserver": "2001:db8::1:4"}}


def test_mine_get_acl_rejected(funcs):
    """
    Asserts that ``mine_get`` gives the expected results when this is rejected
    in the client-side ACL that was stored in the mine data. This results in
    no data being sent back (just as if the entry wouldn't exist).
    """
    funcs.cache.store(
        "minions/webserver",
        "mine",
        {
            "ip_addr": {
                salt.utils.mine.MINE_ITEM_ACL_DATA: "2001:db8::1:4",
                salt.utils.mine.MINE_ITEM_ACL_ID: salt.utils.mine.MINE_ITEM_ACL_VERSION,
                "allow_tgt": "not_requester_minion",
                "allow_tgt_type": "glob",
            }
        },
    )
    # The glob check is for the resolution of the allow_tgt
    # The compound check is for the resolution of the tgt in the mine_get request.
    with patch(
        "salt.utils.minions.CkMinions._check_glob_minions",
        MagicMock(return_value={"minions": ["not_requester_minion"], "missing": []}),
    ), patch(
        "salt.utils.minions.CkMinions._check_compound_minions",
        MagicMock(return_value={"minions": ["webserver"], "missing": []}),
    ):
        ret = funcs._mine_get(
            {
                "id": "requester_minion",
                "tgt": "anything",
                "tgt_type": "compound",
                "fun": ["ip_addr"],
            }
        )
    assert ret == {}



"""
s6 service module

This module is compatible with the :mod:`service <salt.states.service>` states,
so it can be used to maintain services using the ``provider`` argument:

.. code-block:: yaml

    myservice:
      service:
        - running
        - provider: s6

Note that the ``enabled`` argument is not available with this provider.

:codeauthor: Marek Skrobacki <skrobul@skrobul.com>
"""

import os
import re

from salt.exceptions import CommandExecutionError

__func_alias__ = {"reload_": "reload"}

VALID_SERVICE_DIRS = [
    "/service",
    "/etc/service",
]
SERVICE_DIR = None
for service_dir in VALID_SERVICE_DIRS:
    if os.path.exists(service_dir):
        SERVICE_DIR = service_dir
        break


def _service_path(name):
    """
    build service path
    """
    if not SERVICE_DIR:
        raise CommandExecutionError("Could not find service directory.")
    return f"{SERVICE_DIR}/{name}"


def start(name):
    """
    Starts service via s6

    CLI Example:

    .. code-block:: bash

        salt '*' s6.start <service name>
    """
    cmd = f"s6-svc -u {_service_path(name)}"
    return not __salt__["cmd.retcode"](cmd)


def stop(name):
    """
    Stops service via s6

    CLI Example:

    .. code-block:: bash

        salt '*' s6.stop <service name>
    """
    cmd = f"s6-svc -d {_service_path(name)}"
    return not __salt__["cmd.retcode"](cmd)


def term(name):
    """
    Send a TERM to service via s6

    CLI Example:

    .. code-block:: bash

        salt '*' s6.term <service name>
    """
    cmd = f"s6-svc -t {_service_path(name)}"
    return not __salt__["cmd.retcode"](cmd)


def reload_(name):
    """
    Send a HUP to service via s6

    CLI Example:

    .. code-block:: bash

        salt '*' s6.reload <service name>
    """
    cmd = f"s6-svc -h {_service_path(name)}"
    return not __salt__["cmd.retcode"](cmd)


def restart(name):
    """
    Restart service via s6. This will stop/start service

    CLI Example:

    .. code-block:: bash

        salt '*' s6.restart <service name>
    """
    cmd = f"s6-svc -t {_service_path(name)}"
    return not __salt__["cmd.retcode"](cmd)


def full_restart(name):
    """
    Calls s6.restart() function

    CLI Example:

    .. code-block:: bash

        salt '*' s6.full_restart <service name>
    """
    restart(name)


def status(name, sig=None):
    """
    Return the status for a service via s6, return pid if running

    CLI Example:

    .. code-block:: bash

        salt '*' s6.status <service name>
    """
    cmd = f"s6-svstat {_service_path(name)}"
    out = __salt__["cmd.run_stdout"](cmd)
    try:
        pid = re.search(r"up \(pid (\d+)\)", out).group(1)
    except AttributeError:
        pid = ""
    return pid


def available(name):
    """
    Returns ``True`` if the specified service is available, otherwise returns
    ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' s6.available foo
    """
    return name in get_all()


def missing(name):
    """
    The inverse of s6.available.
    Returns ``True`` if the specified service is not available, otherwise returns
    ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' s6.missing foo
    """
    return name not in get_all()


def get_all():
    """
    Return a list of all available services

    CLI Example:

    .. code-block:: bash

        salt '*' s6.get_all
    """
    if not SERVICE_DIR:
        raise CommandExecutionError("Could not find service directory.")
    service_list = [
        dirname for dirname in os.listdir(SERVICE_DIR) if not dirname.startswith(".")
    ]
    return sorted(service_list)



import salt.config
from tests.support.mock import MagicMock, patch


def test_apply_no_cluster_id():
    defaults = salt.config.DEFAULT_MASTER_OPTS.copy()
    assert defaults["cluster_id"] is None

    overrides = {}

    opts = salt.config.apply_master_config(overrides, defaults)
    assert "cluster_id" in opts
    assert opts["cluster_id"] is None
    assert "cluster_pki_dir" in opts
    assert opts["cluster_pki_dir"] is None
    assert "cluster_pool_port" in opts
    assert opts["cluster_pool_port"] == 4520


def test_apply_default_for_cluster():
    defaults = salt.config.DEFAULT_MASTER_OPTS.copy()
    assert defaults["cluster_id"] is None

    overrides = {"cluster_id": "test-cluster"}

    opts = salt.config.apply_master_config(overrides, defaults)
    assert "cluster_id" in opts
    assert "test-cluster" == opts["cluster_id"]

    # the cluster pki dir defaults to pki_dir
    assert "cluster_pki_dir" in opts
    assert opts["pki_dir"] == opts["cluster_pki_dir"]

    # the cluster peers defaults to empty list
    assert "cluster_peers" in opts
    assert [] == opts["cluster_peers"]

    # the cluster pool port defaults to 4520
    assert "cluster_pool_port" in opts
    assert opts["cluster_pool_port"] == 4520


def test_apply_for_cluster():
    defaults = salt.config.DEFAULT_MASTER_OPTS.copy()
    assert defaults["cluster_id"] is None

    cluster_dir = "/tmp/cluster"
    overrides = {
        "cluster_id": "test-cluster",
        "cluster_peers": [
            "127.0.0.1",
            "127.0.0.3",
        ],
        "cluster_pki_dir": cluster_dir,
        "cluster_pool_port": 5500,
    }

    opts = salt.config.apply_master_config(overrides, defaults)
    assert "cluster_id" in opts
    assert "test-cluster" == opts["cluster_id"]

    # the cluster pki dir defaults to pki_dir
    assert "cluster_pki_dir" in opts
    assert cluster_dir == opts["cluster_pki_dir"]

    # the cluster pool port defaults to 4520
    assert "cluster_pool_port" in opts
    assert opts["cluster_pool_port"] == 5500

    # the cluster peers defaults to empty list
    assert "cluster_peers" in opts
    assert isinstance(opts["cluster_peers"], list)
    opts["cluster_peers"].sort()
    assert ["127.0.0.1", "127.0.0.3"] == opts["cluster_peers"]


def test___cli_path_is_expanded():
    defaults = salt.config.DEFAULT_MASTER_OPTS.copy()
    overrides = {}
    with patch(
        "salt.utils.path.expand", MagicMock(return_value="/path/to/testcli")
    ) as expand_mock:
        opts = salt.config.apply_master_config(overrides, defaults)
        assert expand_mock.called
        assert opts["__cli"] == "testcli"



"""
    tests.pytests.unit.beacons.test_haproxy
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    HAProxy beacon test cases
"""

import pytest

import salt.beacons.haproxy as haproxy
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {haproxy: {"__context__": {}, "__salt__": {}}}


def test_non_list_config():
    config = {}

    ret = haproxy.validate(config)
    assert ret == (False, "Configuration for haproxy beacon must be a list.")


def test_empty_config():
    config = [{}]

    ret = haproxy.validate(config)
    assert ret == (False, "Configuration for haproxy beacon requires backends.")


def test_no_servers():
    config = [{"backends": {"www-backend": {"threshold": 45}}}]

    ret = haproxy.validate(config)
    assert ret == (False, "Backends for haproxy beacon require servers.")


def test_threshold_reached():
    config = [{"backends": {"www-backend": {"threshold": 45, "servers": ["web1"]}}}]

    ret = haproxy.validate(config)
    assert ret == (True, "Valid beacon configuration")

    mock = MagicMock(return_value=46)
    with patch.dict(haproxy.__salt__, {"haproxy.get_sessions": mock}):
        ret = haproxy.beacon(config)
        assert ret == [{"threshold": 45, "scur": 46, "server": "web1"}]


def test_threshold_not_reached():
    config = [{"backends": {"www-backend": {"threshold": 100, "servers": ["web1"]}}}]

    ret = haproxy.validate(config)
    assert ret == (True, "Valid beacon configuration")

    mock = MagicMock(return_value=50)
    with patch.dict(haproxy.__salt__, {"haproxy.get_sessions": mock}):
        ret = haproxy.beacon(config)
        assert ret == []



"""
Icinga2 Common Utils
=================

This module provides common functionality for icinga2 module and state.

.. versionadded:: 2018.8.3
"""

import logging
import re

import salt.modules.cmdmod
import salt.utils.path

__salt__ = {"cmd.run_all": salt.modules.cmdmod.run_all}
log = logging.getLogger(__name__)


def get_certs_path():
    icinga2_output = __salt__["cmd.run_all"](
        [salt.utils.path.which("icinga2"), "--version"], python_shell=False
    )
    version = re.search(r"r\d+\.\d+", icinga2_output["stdout"]).group(0)
    # Return new certs path for icinga2 >= 2.8
    if int(version.split(".")[1]) >= 8:
        return "/var/lib/icinga2/certs/"
    # Keep backwords compatibility with older icinga2
    return "/etc/icinga2/pki/"



#
# Copyright 2015 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import os
import time

import salt.utils.files
import salt.utils.fsutils
import salt.utils.network
from salt.modules.inspectlib import EnvLoader
from salt.modules.inspectlib.entities import Package, PackageCfgFile, PayloadFile
from salt.modules.inspectlib.exceptions import InspectorQueryException, SIException

log = logging.getLogger(__name__)


class SysInfo:
    """
    System information.
    """

    def __init__(self, systype):
        if systype.lower() == "solaris":
            raise SIException(f"Platform {systype} not (yet) supported.")

    def _grain(self, grain):
        """
        An alias for grains getter.
        """
        return __grains__.get(grain, "N/A")

    def _get_disk_size(self, device):
        """
        Get a size of a disk.
        """
        out = __salt__["cmd.run_all"](f"df {device}")
        if out["retcode"]:
            msg = "Disk size info error: {}".format(out["stderr"])
            log.error(msg)
            raise SIException(msg)

        devpath, blocks, used, available, used_p, mountpoint = (
            elm for elm in out["stdout"].split(os.linesep)[-1].split(" ") if elm
        )
        return {
            "device": devpath,
            "blocks": blocks,
            "used": used,
            "available": available,
            "used (%)": used_p,
            "mounted": mountpoint,
        }

    def _get_fs(self):
        """
        Get available file systems and their types.
        """

        data = dict()
        for dev, dev_data in salt.utils.fsutils._blkid().items():
            dev = self._get_disk_size(dev)
            device = dev.pop("device")
            dev["type"] = dev_data.get("type", "UNKNOWN")
            data[device] = dev

        return data

    def _get_mounts(self):
        """
        Get mounted FS on the system.
        """
        return salt.utils.fsutils._get_mounts()

    def _get_cpu(self):
        """
        Get available CPU information.
        """
        # CPU data in grains is OK-ish, but lscpu is still better in this case
        out = __salt__["cmd.run_all"]("lscpu")
        salt.utils.fsutils._verify_run(out)
        data = dict()
        for descr, value in [
            elm.split(":", 1) for elm in out["stdout"].split(os.linesep)
        ]:
            data[descr.strip()] = value.strip()

        return data

    def _get_mem(self):
        """
        Get memory.
        """
        out = __salt__["cmd.run_all"]("vmstat -s")
        if out["retcode"]:
            raise SIException("Memory info error: {}".format(out["stderr"]))

        ret = dict()
        for line in out["stdout"].split(os.linesep):
            line = line.strip()
            if not line:
                continue
            size, descr = line.split(" ", 1)
            if descr.startswith("K "):
                descr = descr[2:]
                size = size + "K"
            ret[descr] = size
        return ret

    def _get_network(self):
        """
        Get network configuration.
        """
        data = dict()
        data["interfaces"] = salt.utils.network.interfaces()
        data["subnets"] = salt.utils.network.subnets()

        return data

    def _get_os(self):
        """
        Get operating system summary
        """
        return {
            "name": self._grain("os"),
            "family": self._grain("os_family"),
            "arch": self._grain("osarch"),
            "release": self._grain("osrelease"),
        }


class Query(EnvLoader):
    """
    Query the system.
    This class is actually puts all Salt features together,
    so there would be no need to pick it from various places.
    """

    # Configuration: config files
    # Identity: users/groups
    # Software: packages, patterns, repositories
    # Services
    # System: distro, RAM etc
    # Changes: all files that are managed and were changed from the original
    # all: include all scopes (scary!)
    # payload: files that are not managed

    SCOPES = [
        "changes",
        "configuration",
        "identity",
        "system",
        "software",
        "services",
        "payload",
        "all",
    ]

    def __init__(self, scope, cachedir=None):
        """
        Constructor.

        :param scope:
        :return:
        """
        if scope and scope not in self.SCOPES:
            raise InspectorQueryException(
                "Unknown scope: {}. Must be one of: {}".format(
                    repr(scope), ", ".join(self.SCOPES)
                )
            )
        elif not scope:
            raise InspectorQueryException(
                "Scope cannot be empty. Must be one of: {}".format(
                    ", ".join(self.SCOPES)
                )
            )
        EnvLoader.__init__(self, cachedir=cachedir)
        self.scope = "_" + scope
        self.local_identity = dict()

    def __call__(self, *args, **kwargs):
        """
        Call the query with the defined scope.

        :param args:
        :param kwargs:
        :return:
        """

        return getattr(self, self.scope)(*args, **kwargs)

    def _changes(self, *args, **kwargs):
        """
        Returns all diffs to the configuration files.
        """
        raise Exception("Not yet implemented")

    def _configuration(self, *args, **kwargs):
        """
        Return configuration files.
        """

        data = dict()
        self.db.open()
        for pkg in self.db.get(Package):
            configs = list()
            for pkg_cfg in self.db.get(PackageCfgFile, eq={"pkgid": pkg.id}):
                configs.append(pkg_cfg.path)
            data[pkg.name] = configs

        if not data:
            raise InspectorQueryException("No inspected configuration yet available.")

        return data

    def _get_local_users(self, disabled=None):
        """
        Return all known local accounts to the system.
        """
        users = dict()
        path = "/etc/passwd"
        with salt.utils.files.fopen(path, "r") as fp_:
            for line in fp_:
                line = line.strip()
                if ":" not in line:
                    continue
                name, password, uid, gid, gecos, directory, shell = line.split(":")
                active = not (password == "*" or password.startswith("!"))
                if (
                    (disabled is False and active)
                    or (disabled is True and not active)
                    or disabled is None
                ):
                    users[name] = {
                        "uid": uid,
                        "git": gid,
                        "info": gecos,
                        "home": directory,
                        "shell": shell,
                        "disabled": not active,
                    }

        return users

    def _get_local_groups(self):
        """
        Return all known local groups to the system.
        """
        groups = dict()
        path = "/etc/group"
        with salt.utils.files.fopen(path, "r") as fp_:
            for line in fp_:
                line = line.strip()
                if ":" not in line:
                    continue
                name, password, gid, users = line.split(":")
                groups[name] = {
                    "gid": gid,
                }

                if users:
                    groups[name]["users"] = users.split(",")

        return groups

    def _get_external_accounts(self, locals):
        """
        Return all known accounts, excluding local accounts.
        """
        users = dict()
        out = __salt__["cmd.run_all"]("passwd -S -a")
        if out["retcode"]:
            # System does not supports all accounts descriptions, just skipping.
            return users
        status = {
            "L": "Locked",
            "NP": "No password",
            "P": "Usable password",
            "LK": "Locked",
        }
        for data in [
            elm.strip().split(" ")
            for elm in out["stdout"].split(os.linesep)
            if elm.strip()
        ]:
            if len(data) < 2:
                continue
            name, login = data[:2]
            if name not in locals:
                users[name] = {"login": login, "status": status.get(login, "N/A")}

        return users

    def _identity(self, *args, **kwargs):
        """
        Local users and groups.

        accounts
            Can be either 'local', 'remote' or 'all' (equal to "local,remote").
            Remote accounts cannot be resolved on all systems, but only
            those, which supports 'passwd -S -a'.

        disabled
            True (or False, default) to return only disabled accounts.
        """
        LOCAL = "local accounts"
        EXT = "external accounts"

        data = dict()
        data[LOCAL] = self._get_local_users(disabled=kwargs.get("disabled"))
        data[EXT] = self._get_external_accounts(data[LOCAL].keys()) or "N/A"
        data["local groups"] = self._get_local_groups()

        return data

    def _system(self, *args, **kwargs):
        """
        This basically calls grains items and picks out only
        necessary information in a certain structure.

        :param args:
        :param kwargs:
        :return:
        """
        sysinfo = SysInfo(__grains__.get("kernel"))

        data = dict()
        data["cpu"] = sysinfo._get_cpu()
        data["disks"] = sysinfo._get_fs()
        data["mounts"] = sysinfo._get_mounts()
        data["memory"] = sysinfo._get_mem()
        data["network"] = sysinfo._get_network()
        data["os"] = sysinfo._get_os()

        return data

    def _software(self, *args, **kwargs):
        """
        Return installed software.
        """
        data = dict()
        if "exclude" in kwargs:
            excludes = kwargs["exclude"].split(",")
        else:
            excludes = list()

        os_family = __grains__.get("os_family").lower()

        # Get locks
        if os_family == "suse":
            LOCKS = "pkg.list_locks"
            if "products" not in excludes:
                products = __salt__["pkg.list_products"]()
                if products:
                    data["products"] = products
        elif os_family == "redhat":
            LOCKS = "pkg.get_locked_packages"
        else:
            LOCKS = None

        if LOCKS and "locks" not in excludes:
            locks = __salt__[LOCKS]()
            if locks:
                data["locks"] = locks

        # Get patterns
        if os_family == "suse":
            PATTERNS = "pkg.list_installed_patterns"
        elif os_family == "redhat":
            PATTERNS = "pkg.group_list"
        else:
            PATTERNS = None

        if PATTERNS and "patterns" not in excludes:
            patterns = __salt__[PATTERNS]()
            if patterns:
                data["patterns"] = patterns

        # Get packages
        if "packages" not in excludes:
            data["packages"] = __salt__["pkg.list_pkgs"]()

        # Get repositories
        if "repositories" not in excludes:
            repos = __salt__["pkg.list_repos"]()
            if repos:
                data["repositories"] = repos

        return data

    def _services(self, *args, **kwargs):
        """
        Get list of enabled and disabled services on the particular system.
        """
        return {
            "enabled": __salt__["service.get_enabled"](),
            "disabled": __salt__["service.get_disabled"](),
        }

    def _id_resolv(self, iid, named=True, uid=True):
        """
        Resolve local users and groups.

        :param iid:
        :param named:
        :param uid:
        :return:
        """

        if not self.local_identity:
            self.local_identity["users"] = self._get_local_users()
            self.local_identity["groups"] = self._get_local_groups()

        if not named:
            return iid

        for name, meta in self.local_identity[uid and "users" or "groups"].items():
            if (uid and int(meta.get("uid", -1)) == iid) or (
                not uid and int(meta.get("gid", -1)) == iid
            ):
                return name

        return iid

    def _payload(self, *args, **kwargs):
        """
        Find all unmanaged files. Returns maximum 1000 values.

        Parameters:

        * **filter**: Include only results which path starts from the filter string.
        * **time**: Display time in Unix ticks or format according to the configured TZ (default)
                    Values: ticks, tz (default)
        * **size**: Format size. Values: B, KB, MB, GB
        * **owners**: Resolve UID/GID to an actual names or leave them numeric (default).
                      Values: name (default), id
        * **type**: Comma-separated type of included payload: dir (or directory), link and/or file.
        * **brief**: Return just a list of matches, if True. Default: False
        * **offset**: Offset of the files
        * **max**: Maximum returned values. Default 1000.

        Options:

        * **total**: Return a total amount of found payload files
        """

        def _size_format(size, fmt):
            if fmt is None:
                return size

            fmt = fmt.lower()
            if fmt == "b":
                return f"{size} Bytes"
            elif fmt == "kb":
                return f"{round((float(size) / 0x400), 2)} Kb"
            elif fmt == "mb":
                return f"{round((float(size) / 0x400 / 0x400), 2)} Mb"
            elif fmt == "gb":
                return f"{round((float(size) / 0x400 / 0x400 / 0x400), 2)} Gb"

        filter = kwargs.get("filter")
        offset = kwargs.get("offset", 0)

        timeformat = kwargs.get("time", "tz")
        if timeformat not in ["ticks", "tz"]:
            raise InspectorQueryException(
                f'Unknown "{timeformat}" value for parameter "time"'
            )

        def tfmt(param):
            return (
                timeformat == "tz"
                and time.strftime("%b %d %Y %H:%M:%S", time.gmtime(param))
                or int(param)
            )

        size_fmt = kwargs.get("size")
        if size_fmt is not None and size_fmt.lower() not in ["b", "kb", "mb", "gb"]:
            raise InspectorQueryException(
                'Unknown "{}" value for parameter "size". '
                "Should be either B, Kb, Mb or Gb".format(timeformat)
            )

        owners = kwargs.get("owners", "id")
        if owners not in ["name", "id"]:
            raise InspectorQueryException(
                'Unknown "{}" value for parameter "owners". '
                "Should be either name or id (default)".format(owners)
            )

        incl_type = [prm for prm in kwargs.get("type", "").lower().split(",") if prm]
        if not incl_type:
            incl_type.append("file")

        for i_type in incl_type:
            if i_type not in ["directory", "dir", "d", "file", "f", "link", "l"]:
                raise InspectorQueryException(
                    'Unknown "{}" values for parameter "type". '
                    "Should be comma separated one or more of "
                    "dir, file and/or link.".format(", ".join(incl_type))
                )
        self.db.open()

        if "total" in args:
            return {"total": len(self.db.get(PayloadFile))}

        brief = kwargs.get("brief")
        pld_files = list() if brief else dict()
        for pld_data in self.db.get(PayloadFile)[
            offset : offset + kwargs.get("max", 1000)
        ]:
            if brief:
                pld_files.append(pld_data.path)
            else:
                pld_files[pld_data.path] = {
                    "uid": self._id_resolv(pld_data.uid, named=owners == "id"),
                    "gid": self._id_resolv(
                        pld_data.gid, named=owners == "id", uid=False
                    ),
                    "size": _size_format(pld_data.p_size, fmt=size_fmt),
                    "mode": oct(pld_data.mode),
                    "accessed": tfmt(pld_data.atime),
                    "modified": tfmt(pld_data.mtime),
                    "created": tfmt(pld_data.ctime),
                }

        return pld_files

    def _all(self, *args, **kwargs):
        """
        Return all the summary of the particular system.
        """
        data = dict()
        data["software"] = self._software(**kwargs)
        data["system"] = self._system(**kwargs)
        data["services"] = self._services(**kwargs)
        try:
            data["configuration"] = self._configuration(**kwargs)
        except InspectorQueryException as ex:
            data["configuration"] = "N/A"
            log.error(ex)
        data["payload"] = self._payload(**kwargs) or "N/A"

        return data



import logging
import random
import string

import pytest

import salt.loader
import salt.states.boto_cloudtrail as boto_cloudtrail
from tests.support.mock import MagicMock, patch

boto = pytest.importorskip("boto")
boto3 = pytest.importorskip("boto3", "1.2.1")
botocore = pytest.importorskip("botocore", "1.4.41")

log = logging.getLogger(__name__)


class GlobalConfig:
    region = "us-east-1"
    access_key = "GKTADJGHEIQSXMKKRBJ08H"
    secret_key = "askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs"
    conn_parameters = {
        "region": region,
        "key": access_key,
        "keyid": secret_key,
        "profile": {},
    }
    error_message = (
        "An error occurred (101) when calling the {0} operation: Test-defined error"
    )
    not_found_error = botocore.exceptions.ClientError(
        {"Error": {"Code": "TrailNotFoundException", "Message": "Test-defined error"}},
        "msg",
    )
    error_content = {"Error": {"Code": 101, "Message": "Test-defined error"}}
    trail_ret = dict(
        Name="testtrail",
        IncludeGlobalServiceEvents=True,
        KmsKeyId=None,
        LogFileValidationEnabled=False,
        S3BucketName="auditinfo",
        TrailARN="arn:aws:cloudtrail:us-east-1:214351231622:trail/testtrail",
    )
    status_ret = dict(
        IsLogging=False,
        LatestCloudWatchLogsDeliveryError=None,
        LatestCloudWatchLogsDeliveryTime=None,
        LatestDeliveryError=None,
        LatestDeliveryTime=None,
        LatestDigestDeliveryError=None,
        LatestDigestDeliveryTime=None,
        LatestNotificationError=None,
        LatestNotificationTime=None,
        StartLoggingTime=None,
        StopLoggingTime=None,
    )


@pytest.fixture
def session_instance():
    with patch("boto3.session.Session") as patched_session:
        yield patched_session()


@pytest.fixture
def global_config():
    GlobalConfig.conn_parameters["key"] = "".join(
        random.choice(string.ascii_lowercase + string.digits) for _ in range(50)
    )
    params = GlobalConfig()
    return params


@pytest.fixture
def configure_loader_modules(minion_opts):
    minion_opts["grains"] = salt.loader.grains(minion_opts)
    ctx = {}
    utils = salt.loader.utils(
        minion_opts,
        whitelist=["boto", "boto3", "args", "systemd", "path", "platform", "reg"],
        context=ctx,
    )
    serializers = salt.loader.serializers(minion_opts)
    funcs = salt.loader.minion_mods(
        minion_opts, context=ctx, utils=utils, whitelist=["boto_cloudtrail"]
    )
    salt_states = salt.loader.states(
        opts=minion_opts,
        functions=funcs,
        utils=utils,
        whitelist=["boto_cloudtrail"],
        serializers=serializers,
    )
    return {
        boto_cloudtrail: {
            "__opts__": minion_opts,
            "__salt__": funcs,
            "__utils__": utils,
            "__states__": salt_states,
            "__serializers__": serializers,
        }
    }


@pytest.mark.slow_test
def test_present_when_trail_does_not_exist(global_config, session_instance):
    """
    Tests present on a trail that does not exist.
    """
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.get_trail_status.side_effect = [
        global_config.not_found_error,
        global_config.status_ret,
    ]
    conn.create_trail.return_value = global_config.trail_ret
    conn.describe_trails.return_value = {"trailList": [global_config.trail_ret]}
    with patch.dict(
        boto_cloudtrail.__salt__,
        {"boto_iam.get_account_id": MagicMock(return_value="1234")},
    ):
        result = boto_cloudtrail.__states__["boto_cloudtrail.present"](
            "trail present",
            Name=global_config.trail_ret["Name"],
            S3BucketName=global_config.trail_ret["S3BucketName"],
        )

    assert result["result"]
    assert result["changes"]["new"]["trail"]["Name"] == global_config.trail_ret["Name"]


@pytest.mark.slow_test
def test_present_when_trail_exists(global_config, session_instance):
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.get_trail_status.return_value = global_config.status_ret
    conn.create_trail.return_value = global_config.trail_ret
    conn.describe_trails.return_value = {"trailList": [global_config.trail_ret]}
    with patch.dict(
        boto_cloudtrail.__salt__,
        {"boto_iam.get_account_id": MagicMock(return_value="1234")},
    ):
        result = boto_cloudtrail.__states__["boto_cloudtrail.present"](
            "trail present",
            Name=global_config.trail_ret["Name"],
            S3BucketName=global_config.trail_ret["S3BucketName"],
            LoggingEnabled=False,
        )
    assert result["result"]
    assert result["changes"] == {}


@pytest.mark.slow_test
def test_present_with_failure(global_config, session_instance):
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.get_trail_status.side_effect = [
        global_config.not_found_error,
        global_config.status_ret,
    ]
    conn.create_trail.side_effect = botocore.exceptions.ClientError(
        global_config.error_content, "create_trail"
    )
    with patch.dict(
        boto_cloudtrail.__salt__,
        {"boto_iam.get_account_id": MagicMock(return_value="1234")},
    ):
        result = boto_cloudtrail.__states__["boto_cloudtrail.present"](
            "trail present",
            Name=global_config.trail_ret["Name"],
            S3BucketName=global_config.trail_ret["S3BucketName"],
            LoggingEnabled=False,
        )
    assert not result["result"]
    assert "An error occurred" in result["comment"]


def test_absent_when_trail_does_not_exist(global_config, session_instance):
    """
    Tests absent on a trail that does not exist.
    """
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.get_trail_status.side_effect = global_config.not_found_error
    result = boto_cloudtrail.__states__["boto_cloudtrail.absent"]("test", "mytrail")
    assert result["result"]
    assert result["changes"] == {}


def test_absent_when_trail_exists(global_config, session_instance):
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.get_trail_status.return_value = global_config.status_ret
    result = boto_cloudtrail.__states__["boto_cloudtrail.absent"](
        "test", global_config.trail_ret["Name"]
    )
    assert result["result"]
    assert result["changes"]["new"]["trail"] is None


def test_absent_with_failure(global_config, session_instance):
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.get_trail_status.return_value = global_config.status_ret
    conn.delete_trail.side_effect = botocore.exceptions.ClientError(
        global_config.error_content, "delete_trail"
    )
    result = boto_cloudtrail.__states__["boto_cloudtrail.absent"](
        "test", global_config.trail_ret["Name"]
    )
    assert not result["result"]
    assert "An error occurred" in result["comment"]



"""
Management of Microsoft SQLServer Users
=======================================

The mssql_user module is used to create
and manage SQL Server Users

.. code-block:: yaml

    frank:
      mssql_user.present:
        - database: yolo
"""

import collections


def __virtual__():
    """
    Only load if the mssql module is present
    """
    if "mssql.version" in __salt__:
        return True
    return (False, "mssql module could not be loaded")


def _normalize_options(options):
    if type(options) in [dict, collections.OrderedDict]:
        return [f"{k}={v}" for k, v in options.items()]
    if type(options) is list and (not options or type(options[0]) is str):
        return options
    # Invalid options
    if type(options) is not list or type(options[0]) not in [
        dict,
        collections.OrderedDict,
    ]:
        return []
    return [o for d in options for o in _normalize_options(d)]


def present(
    name, login=None, domain=None, database=None, roles=None, options=None, **kwargs
):
    """
    Checks existence of the named user.
    If not present, creates the user with the specified roles and options.

    name
        The name of the user to manage
    login
        If not specified, will be created WITHOUT LOGIN
    domain
        Creates a Windows authentication user.
        Needs to be NetBIOS domain or hostname
    database
        The database of the user (not the login)
    roles
        Add this user to all the roles in the list
    options
        Can be a list of strings, a dictionary, or a list of dictionaries
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    if domain and not login:
        ret["result"] = False
        ret["comment"] = "domain cannot be set without login"
        return ret
    if __salt__["mssql.user_exists"](name, domain=domain, database=database, **kwargs):
        ret["comment"] = (
            "User {} is already present (Not going to try to set its roles or options)".format(
                name
            )
        )
        return ret
    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = f"User {name} is set to be added"
        return ret

    user_created = __salt__["mssql.user_create"](
        name,
        login=login,
        domain=domain,
        database=database,
        roles=roles,
        options=_normalize_options(options),
        **kwargs,
    )
    if (
        user_created is not True
    ):  # Non-empty strings are also evaluated to True, so we cannot use if not user_created:
        ret["result"] = False
        ret["comment"] += f"User {name} failed to be added: {user_created}"
        return ret
    ret["comment"] += f"User {name} has been added"
    ret["changes"][name] = "Present"
    return ret


def absent(name, **kwargs):
    """
    Ensure that the named user is absent

    name
        The username of the user to remove
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    if not __salt__["mssql.user_exists"](name):
        ret["comment"] = f"User {name} is not present"
        return ret
    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = f"User {name} is set to be removed"
        return ret
    if __salt__["mssql.user_remove"](name, **kwargs):
        ret["comment"] = f"User {name} has been removed"
        ret["changes"][name] = "Absent"
        return ret
    # else:
    ret["result"] = False
    ret["comment"] = f"User {name} failed to be removed"
    return ret



"""
IRC Bot engine

.. versionadded:: 2017.7.0

Example Configuration

.. code-block:: yaml

    engines:
      - ircbot:
          nick: <nick>
          username: <username>
          password: <password>
          host: irc.oftc.net
          port: 7000
          channels:
            - salt-test
            - '##something'
          use_ssl: True
          use_sasl: True
          disable_query: True
          allow_hosts:
            - salt/engineer/.*
          allow_nicks:
            - gtmanfred

Available commands on irc are:

ping
    return pong

echo <stuff>
    return <stuff> targeted at the user who sent the commands

event <tag> [<extra>, <data>]
    fire event on the master or minion event stream with the tag `salt/engines/ircbot/<tag>` and a data object with a
    list of everything else sent in the message

Example of usage

.. code-block:: text

    08:33:57 @gtmanfred > !ping
    08:33:57   gtmanbot > gtmanfred: pong
    08:34:02 @gtmanfred > !echo ping
    08:34:02   gtmanbot > ping
    08:34:17 @gtmanfred > !event test/tag/ircbot irc is useful
    08:34:17   gtmanbot > gtmanfred: TaDa!

.. code-block:: text

    [DEBUG   ] Sending event: tag = salt/engines/ircbot/test/tag/ircbot; data = {'_stamp': '2016-11-28T14:34:16.633623', 'data': ['irc', 'is', 'useful']}

"""

import base64
import logging
import re
import socket
import ssl
from collections import namedtuple

import tornado.ioloop
import tornado.iostream

import salt.utils.event

log = logging.getLogger(__name__)


# Nothing listening here
Event = namedtuple("Event", "source code line")
PrivEvent = namedtuple("PrivEvent", "source nick user host code channel command line")


class IRCClient:
    def __init__(
        self,
        nick,
        host,
        port=6667,
        username=None,
        password=None,
        channels=None,
        use_ssl=False,
        use_sasl=False,
        char="!",
        allow_hosts=False,
        allow_nicks=False,
        disable_query=True,
    ):
        self.nick = nick
        self.host = host
        self.port = port
        self.username = username or nick
        self.password = password
        self.channels = channels or []
        self.ssl = use_ssl
        self.sasl = use_sasl
        self.char = char
        self.allow_hosts = allow_hosts
        self.allow_nicks = allow_nicks
        self.disable_query = disable_query
        self.io_loop = tornado.ioloop.IOLoop()
        self._connect()

    def _connect(self):
        _sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
        if self.ssl is True:
            self._stream = tornado.iostream.SSLIOStream(
                _sock, ssl_options={"cert_reqs": ssl.CERT_NONE}
            )
        else:
            self._stream = tornado.iostream.IOStream(_sock)
        self._stream.set_close_callback(self.on_closed)
        self._stream.connect((self.host, self.port), self.on_connect)

    def read_messages(self):
        self._stream.read_until("\r\n", self._message)

    @staticmethod
    def _event(line):
        log.debug("Received: %s", line)
        search = re.match(
            "^(?:(?P<source>:[^ ]+) )?(?P<code>[^ ]+)(?: (?P<line>.*))?$", line
        )
        source, code, line = (
            search.group("source"),
            search.group("code"),
            search.group("line"),
        )
        return Event(source, code, line)

    def _allow_host(self, host):
        if isinstance(self.allow_hosts, bool):
            return self.allow_hosts
        else:
            return any([re.match(match, host) for match in self.allow_hosts])

    def _allow_nick(self, nick):
        if isinstance(self.allow_nicks, bool):
            return self.allow_nicks
        else:
            return any([re.match(match, nick) for match in self.allow_nicks])

    def _privmsg(self, event):
        search = re.match(
            "^:(?P<nick>[^!]+)!(?P<user>[^@]+)@(?P<host>.*)$", event.source
        )
        nick, user, host = (
            search.group("nick"),
            search.group("user"),
            search.group("host"),
        )
        search = re.match(
            "^(?P<channel>[^ ]+) :(?:{}(?P<command>[^ ]+)(?: (?P<line>.*))?)?$".format(
                self.char
            ),
            event.line,
        )
        if search:
            channel, command, line = (
                search.group("channel"),
                search.group("command"),
                search.group("line"),
            )
            if self.disable_query is True and not channel.startswith("#"):
                return
            if channel == self.nick:
                channel = nick
            privevent = PrivEvent(
                event.source, nick, user, host, event.code, channel, command, line
            )
            if (self._allow_nick(nick) or self._allow_host(host)) and hasattr(
                self, f"_command_{command}"
            ):
                getattr(self, f"_command_{command}")(privevent)

    def _command_echo(self, event):
        message = f"PRIVMSG {event.channel} :{event.line}"
        self.send_message(message)

    def _command_ping(self, event):
        message = f"PRIVMSG {event.channel} :{event.nick}: pong"
        self.send_message(message)

    def _command_event(self, event):
        if __opts__.get("__role") == "master":
            fire_master = salt.utils.event.get_master_event(
                __opts__, __opts__["sock_dir"]
            ).fire_event
        else:
            fire_master = None

        def fire(tag, msg):
            """
            How to fire the event
            """
            if fire_master:
                fire_master(msg, tag)
            else:
                __salt__["event.send"](tag, msg)

        args = event.line.split(" ")
        tag = args[0]
        if len(args) > 1:
            payload = {"data": args[1:]}
        else:
            payload = {"data": []}

        fire("salt/engines/ircbot/" + tag, payload)
        message = f"PRIVMSG {event.channel} :{event.nick}: TaDa!"
        self.send_message(message)

    def _message(self, raw):
        raw = raw.rstrip(b"\r\n").decode("utf-8")
        event = self._event(raw)

        if event.code == "PING":
            tornado.ioloop.IOLoop.current().spawn_callback(
                self.send_message, f"PONG {event.line}"
            )
        elif event.code == "PRIVMSG":
            tornado.ioloop.IOLoop.current().spawn_callback(self._privmsg, event)
        self.read_messages()

    def join_channel(self, channel):
        if not channel.startswith("#"):
            channel = "#" + channel
        self.send_message(f"JOIN {channel}")

    def on_connect(self):
        logging.info("on_connect")
        if self.sasl is True:
            self.send_message("CAP REQ :sasl")
        self.send_message(f"NICK {self.nick}")
        self.send_message("USER saltstack 0 * :saltstack")
        if self.password:
            if self.sasl is True:
                authstring = base64.b64encode(
                    "{0}\x00{0}\x00{1}".format(self.username, self.password).encode()
                )
                self.send_message("AUTHENTICATE PLAIN")
                self.send_message(f"AUTHENTICATE {authstring}")
                self.send_message("CAP END")
            else:
                self.send_message(
                    "PRIVMSG NickServ :IDENTIFY {} {}".format(
                        self.username, self.password
                    )
                )
        for channel in self.channels:
            self.join_channel(channel)
        self.read_messages()

    def on_closed(self):
        logging.info("on_closed")

    def send_message(self, line):
        if isinstance(line, str):
            line = line.encode("utf-8")
        log.debug("Sending:  %s", line)
        self._stream.write(line + b"\r\n")


def start(
    nick,
    host,
    port=6667,
    username=None,
    password=None,
    channels=None,
    use_ssl=False,
    use_sasl=False,
    char="!",
    allow_hosts=False,
    allow_nicks=False,
    disable_query=True,
):
    """
    IRC Bot for interacting with salt.

    nick
        Nickname of the connected Bot.

    host
        irc server (example - irc.oftc.net).

    port
        irc port.  Default: 6667

    password
        password for authenticating.  If not provided, user will not authenticate on the irc server.

    channels
        channels to join.

    use_ssl
        connect to server using ssl. Default: False

    use_sasl
        authenticate using sasl, instead of messaging NickServ. Default: False

        .. note:: This will allow the bot user to be fully authenticated before joining any channels

    char
        command character to look for. Default: !

    allow_hosts
        hostmasks allowed to use commands on the bot.  Default: False
        True to allow all
        False to allow none
        List of regexes to allow matching

    allow_nicks
        Nicks that are allowed to use commands on the bot.  Default: False
        True to allow all
        False to allow none
        List of regexes to allow matching

    disable_query
        Disable commands from being sent through private queries.  Require they be sent to a channel, so that all
        communication can be controlled by access to the channel. Default: True

    .. warning:: Unauthenticated Access to event stream

        This engine sends events calls to the event stream without authenticating them in salt.  Authentication will
        need to be configured and enforced on the irc server or enforced in the irc channel.  The engine only accepts
        commands from channels, so non authenticated users could be banned or quieted in the channel.

        /mode +q $~a  # quiet all users who are not authenticated
        /mode +r      # do not allow unauthenticated users into the channel

        It would also be possible to add a password to the irc channel, or only allow invited users to join.
    """
    client = IRCClient(
        nick,
        host,
        port,
        username,
        password,
        channels or [],
        use_ssl,
        use_sasl,
        char,
        allow_hosts,
        allow_nicks,
        disable_query,
    )
    client.io_loop.start()
